# Predefined Docgen Productions
#
# These productions are either generally useful or require special magic (such 
# as indent codes) that we really don't want users to be using.

# Imports {{{1
from __future__ import print_function, division
from .preferences import DEFAULT_TABSTOP
from .production import (
    Production, newlineCode, tabCode, killCode, indentCode, leaderCode, 
    dedentCode, emptyCode
)
import collections
import re


# Alternates {{{1
class Alternates(Production):
    """Choose between alternate rules.  Use the first that works.

    :param rules: A list of rules.

    >>> from docgen import Alternates

    >>> choice = Alternates([None, 'coffee', 'tea'])
    >>> print str(choice)
    coffee

    """
    def __init__(self, rules):

        # expand all rules and identify the null ones
        # but do not declare a list null if it contains an null member
        rules = self._clean(rules, cullArrays=False)

        if rules is None:
            self.rule = None
            return
        for rule in rules:
            if rule is not None:
                self.rule = rule
                return
        self.rule = None


# Block {{{1
class Block(Production):
    """
    A list where the contents of the list are indented and each item is placed 
    on a separate line.

    :param items:
        A list of rules. Each will be indented and set on its own line.
    :param sep:
        A rule that is added to the end of each item except the last. The actual 
        separator used will be *sep* followed by Newline().
    :param lcap:
        A rule that is precedes the first item. It is not indented.
    :param rcap:
        A rule that is follows the last item. It is not indented.
    :param term:
        A rule that is added to the end of the last item.
    :param ifempty:
        What to do if the argument is None or an empty list. Choices are
            'error': raise a ValueError,
            'discard': return None to contaminate the containing rule,
            'caps': return the end caps surrounding a newline,
            'capsonly': return the end caps surrounding an empty string.
        Anything else simply replaces the empty list.
    :param ifsingle:
        What to do if the argument is a list with only one item.  Choices are
            'caps': print the newline, indent, and end caps,
            'nocaps': print newline and indent, but not the end caps.
            'bare': do not print newline, indent or end caps.

    >>> from docgen import Block, Join

    >>> stmt = Join([
    ...     'if (doit)',
    ...     Block(['t = a;', 'a = b;', 'b = t;'], lcap='begin', rcap='end')
    ... ])
    >>> print str(stmt)
    if (doit) begin
        t = a;
        a = b;
        b = t;
    end

    >>> print(Block([], ifempty='error'))
    Traceback (most recent call last):
    ...
    ValueError: Block must not be empty.

    >>> print(Block([], ifempty='discard'))
    <BLANKLINE>

    >>> print(Block([], ifempty='caps'))
    (
    )

    >>> print(Block([], ifempty='capsonly'))
    ()

    >>> print(Block(['a'], ifsingle='caps'))
    (
        a
    )

    >>> print(Block([], ifempty='...'))
    (
        ...
    )

    >>> print(Block(['a'], ifsingle='nocaps'))
    <BLANKLINE>
        a

    >>> print(Block(['a'], ifsingle='bare'))
    a

    >>> print(Block([], ifempty='<empty>', ifsingle='bare'))
    <empty>

    """
    def __init__(
        self,
        items,
        sep='', lcap='(', rcap=')', term=emptyCode(),
        ifempty='discard',
        ifsingle='caps'
    ):
        if ifempty != 'caps':
            items = DiscardIfEmpty(items)
        if not items:
            if ifempty == 'discard':
                self.add_rule(None)
                return
            elif ifempty == 'error':
                raise ValueError('Block must not be empty.')
            elif ifempty not in ['caps', 'capsonly']:
                items = [ifempty]

        if items and len(items) == 1:
            if ifsingle == 'nocaps':
                lcap = rcap = ''
            elif ifsingle == 'bare':
                self.add_rule(Join(items))
                return

        self.add_rule([
            lcap,
            Optional(
                Indent(
                    Join(
                        items,
                        sep=[sep, Newline()], rcap=term
                    )
                )
            ),
            rcap
        ])

# Cull {{{1
class Cull(Production):
    """Remove any null rules from a list.

    :param rules:
        A list of rules. Each will be examined to determine if it is inactive 
        and if so, it will be removed from the list.
    :param nonempty:
        A boolean. If true, the resulting list is expected to be nonempty. If it 
        is not, the entire production is nulled. DEPRECATED, use 
        DiscardIfEmpty() instead.

    >>> from docgen import Cull, Join

    >>> choices = Join(Cull([None, 'coffee', 'tea']))
    >>> print str(choices)
    coffee tea

    >>> from docgen import Optional, Join, Indent, Lines

    >>> analog = Optional([
    ...     Join(['analog', 'begin']),
    ...     Indent(Lines(Cull(['a = 1;', None, 'c = 3;']))),
    ...     'end'
    ... ])
    >>> print str(analog)
    analog begin
        a = 1;
        c = 3;
    end

    If the argument *nonempty* is true, then None is returned if the resulting 
    list is empty after all the null rules have been culled.

    >>> analog = Optional(Join([
    ...     'analog',
    ...     'begin',
    ...     Indent(Cull([None, None, None], nonempty=True)),
    ...     'end'
    ... ]))
    >>> print str(analog)
    <BLANKLINE>

    """
    def __init__(self, rules, nonempty=True):

        # expand all rules and identify the null ones
        # but do not declare a list null if it contains an null member
        rules = self._clean(rules, cullArrays=False)

        if rules is None:
            self.rule = None
            return
        self.rule = [rule for rule in rules if rule is not None]
        if nonempty and len(self.rule) == 0:
            self.rule = None

    def __len__(self):
        return len(self.rule)

    def __iter__(self):
        return iter(self.rule)


# DiscardIfEmpty {{{1
class DiscardIfEmpty(Production):
    """Convert rule to null if it is an empty list.

    :param rules:
        A list of rules.

    >>> from docgen import DiscardIfEmpty, Join

    >>> choices = Join(DiscardIfEmpty(['coffee', 'tea']))
    >>> nochoices = Join(DiscardIfEmpty([]))
    >>> print str(choices)
    coffee tea
    >>> print str(nochoices)
    <BLANKLINE>

    """
    def __init__(self, rules):
        self.rule = rules if rules else None

    def __len__(self):
        return len(self.rule)

    def __iter__(self):
        return iter(self.rule)


# Indent {{{1
class Indent(Production):
    """Increase the indent level.

    The contents are indented by the specified number of levels. If not 
    specified, the default is indenting one level. Nested indents accumulate.
    By default, a level of indent is 4 spaces, but this can be overridden when 
    using the expand() Production method.

    There are two indenting productions, this Indent and Indented().

    Indent() uses immediate indenting and dedenting. As such, it requires 
    a newline before the indenting and before the dedenting. The leading newline 
    is often provided by the surrounding rules and the trailing newline is often 
    provided by the contents, though can be provided by the Indent production 
    itself. By default Indent provides both leading and trailing newlines even 
    though they are often redundant, because if they are redundant the extra 
    newlines are automatically suppressed. Generally Indent is used with 
    conventional productions.

    :param contents:
        The rule to be indented.
    :param levels:
        An integer that specifies the number of levels to indent.
    :param newlines:
        Specify 'b' (short for both) if newlines should be added before an after 
        contents (default).  Specify 'l' (short for leading) if newline should 
        be added before the contents.  Specify 't' (short for trailing) if 
        newline should be added after the contents.  Specify 'n' (short for 
        none) if newlines should not be added to the contents.

    >>> from docgen import Indent, Production

    >>> whattodo = Production(['if today == doomsday:', Indent('panic')])
    >>> print str(whattodo)
    if today == doomsday:
        panic

    """
    def __init__(self, contents, levels=1, newlines='b'):
        lcap = Newline() if newlines in ['l', 'b'] else emptyCode()
        rcap = Newline() if newlines in ['t', 'b'] else emptyCode()
        self.add_rule([
            lcap,
            indentCode(levels),
            contents,
            rcap,
            dedentCode(levels),
        ])


# Indented {{{1
class Indented(Production):
    """Increase the indent level.

    There are two indenting productions, Indent() and Indented().

    Indented() uses deferred indenting and dedenting. As such, it expects 
    indenting to be provided on the first line and a newline to immediately 
    follow the production. This is generally true in templates. Generally 
    Indented() is used with templates and Indent() is used with individual 
    productions.

    :param contents:
        The rule to be indented.
    :param levels:
        An integer that specifies the number of levels to indent.
    :param newlines:
        Specify 'b' (short for both) if newlines should be added before an after 
        contents (default).  Specify 'l' (short for leading) if newline should 
        be added before the contents.  Specify 't' (short for trailing) if 
        newline should be added after the contents.  Specify 'n' (short for 
        none) if newlines should not be added to the contents.

    >>> from docgen import Indented, Template, Lines, Newline
    >>> from textwrap import dedent

    >>> whattodo = Template(dedent('''\
    ...     if today == doomsday:
    ...         {Indented(Lines(['panic', 'panic some more']))}
    ... '''), symbols=globals())
    >>> print str(whattodo)
    if today == doomsday:
        panic
        panic some more

    """

    def __init__(self, contents, newlines='n', levels=1):
        lcap = Newline() if newlines in ['l', 'b'] else emptyCode()
        rcap = Newline() if newlines in ['t', 'b'] else emptyCode()
        self.add_rule([
            lcap,
            indentCode(levels, deferred=True),
            contents,
            rcap,
            dedentCode(levels, deferred=True),
        ])


# Join {{{1
class Join(Production):
    """Put separators between the members of a list and add end caps.

    :param rules:
        A list of rules. Each will be joined together using the separator to 
        form a single unit.
    :param sep:
        A rule that is placed between each rule in rules.
    :param lcap:
        A rule that is placed before the first rule in rules.
    :param rcap:
        A rule that is placed after the last rule in rules.

    >>> from docgen import Join

    >>> l = Join(['a', 'b', 'c'])
    >>> print str(l)
    a b c

    >>> l = Join(['a', 'b', 'c'], sep=', ', lcap='<', rcap='>')
    >>> print str(l)
    <a, b, c>

    """
    def __init__(self, rules, sep=' ', lcap=emptyCode(), rcap=emptyCode()):

        # expand all rules and identify the null ones
        rules = self._clean(rules)

        if rules is None:
            self.rule = None
            return

        sep = self._clean(sep)
        lcap = self._clean(lcap)
        rcap = self._clean(rcap)

        if rules:
            # perform the join by creating array twice as long as original 
            # filled with the separator, then do an interleaved assignment to 
            # write over every other separator with a rule in the given list.
            self.rule = [lcap] + (2*len(rules)-1)*[sep] + [rcap]
            self.rule[1::2] = rules
        else:
            self.rule = [lcap, rcap]


# Leader {{{1
class Leader(Production):
    """Add a leader to each line.

    :param contents:
        The rule that, when expanded, will have the leader appended to the 
        beginning of each line.
    :param leader:
        The key (a string) for the desired leader. The available leaders are 
        specified in a dictionary to the expand() method when the production is 
        expanded. If a dictionary of leaders will not be provided when 
        expanding, you must specify a key for one of the default leaders, which 
        includes '*' (maps to ' * '), '#' (maps to '# '), and '//' (maps to '// 
        ').

    >>> from docgen import Leader, Production, Newline, Lines, Indented

    >>> class Comment(Production):
    ...     def __init__(self, lines):
    ...         self.add_rule([
    ...             '/*', Newline(),
    ...             Leader(Lines(lines), '*'),
    ...             Newline(), ' */'
    ...         ])

    >>> title = "Leader"
    >>> description = dedent('''\
    ...     Leader adds a fixed string to the beginning of every line.
    ...     If differs in a fundamental way from Join in that it acts on lines, 
    ...     whereas Join acts on items in a list.
    ... ''')
    >>> comment = Comment([title, '', description])
    >>> print str(comment)
    /*
     * Leader
     *
     * Leader adds a fixed string to the beginning of every line.
     * If differs in a fundamental way from Join in that it acts on lines,
     * whereas Join acts on items in a list.
     */

    It is possible to nest both indents and leaders.

    """
    def __init__(self, contents, leader):
        self.add_rule([
            leaderCode(leader), contents, dedentCode(1), Newline()
        ])


# Lines {{{1
class Lines(Production):
    """Put each entry of a list onto its own line.

    :param items:
        A list of rules. Each will be placed on its own line as it is joined 
        together to form a single unit.

    >>> from docgen import Lines

    >>> l = Lines(['coffee', 'tea', '', 'me'])
    >>> print str(l)
    coffee
    tea
    <BLANKLINE>
    me

    """

    def __init__(self, items):
        self.add_rule(Join(items, sep=Newline(), rcap=Newline()))


# Newline {{{1
class Newline(Production):
    """Adds one or more newlines to the output.

    :param blankLines:
        The number of newlines output equals the number needed to produce number 
        of blank lines specified on the constructor.  Any adjacent newline 
        commands are combined with the result being the number of blank lines 
        produced is the largest specified on any of the adjacent commands.

        If instead '!' is passed to the constructor, Newline acts like a
        conventional newline (adjacent newlines are not combined).

    >>> from docgen import Newline, Production

    >>> l = Production(
    ...     ['coffee', Newline(), Newline(), 'tea', Newline(1), 'me']
    ... )
    >>> print str(l)
    coffee
    tea
    <BLANKLINE>
    me

    >>> l = Production(
    ...     ['coffee', Newline('!'), Newline('!'), 'tea', Newline(2), 'me']
    ... )
    >>> print str(l)
    coffee
    <BLANKLINE>
    tea
    <BLANKLINE>
    <BLANKLINE>
    me

    """
    def __init__(self, blankLines=0):
        self.add_rule(newlineCode(blankLines))


# Op {{{1
class Op(Production):
    """An operator with operands.

    Takes a single operator and a list of operands. The list is joined by the 
    operator.

    :param operator:
        A rule that becomes the operator.
    :param operands:
        A list of rules that become the operands. The list is culled of null 
        rules, and if the resulting list is empty, the operator itself becomes 
        null.
    :param group:
        If false, do not delimit the expression. If true, delimit with 
        parentheses. Otherwise a string of two characters or a string with two 
        groups of characters separated by white space should be specified, and 
        the expression will be delimited by those.  group is ignored if there 
        are less than two operands.
    :param multiline:
        A boolean. If true, place each delimiter and operand on a separate line.
    :param squeeze:
        A boolean. If true, do not put spaces around the operator.

    >>> from docgen import Op

    >>> expr = Op('+', ['a', 'b', 'c'])
    >>> print str(expr)
    a + b + c

    >>> expr = Op(':', ['a', 'b', 'c'], group='[]', squeeze=True)
    >>> print str(expr)
    [a:b:c]

    >>> assign = Op('=', [
    ...     'x',
    ...     Op('+', ['a', 'b', 'c'], group=True, multiline=True)
    ... ])
    >>> print str(assign)
    x = (
        a +
        b +
        c
    )

    >>> expr = Op('+', ['a', 'b', None, 'd'])
    >>> print str(expr)
    <BLANKLINE>

    """
    def __init__(
        self, operator, operands, group=None, multiline=False, squeeze=False
    ):
        if group and operands and len(operands) > 1:
            if group is True:
                group = '()'
            if len(group) != 2:
                group = group.split
            assert len(group)
            lcap = group[0]
            rcap = group[1]
        else:
            lcap = emptyCode()
            rcap = emptyCode()
        sep = operator if squeeze else [' ', operator, ' ']
        joiner = Block if multiline else Join

        self.add_rule(joiner(
            operands, sep=sep, lcap=lcap, rcap=rcap
        ))

# Optional {{{1
class Optional(Production):
    """An optional rule.

    The rule is removed and replaced by alt if it resolves to None.

    :param rule:
        The rule that this production evaluates to as long as this rule is 
        active.
    :param alt:
        The rule that this production evaluates to if the above rule is 
        inactive.

    >>> from docgen import Optional

    >>> result = Optional('Mr. Bill', alt='Oh no!')
    >>> print str(result)
    Mr. Bill

    >>> result = Optional(None, alt='Oh no!')
    >>> print str(result)
    Oh no!

    """
    def __init__(self, rule, alt=emptyCode()):
        self.add_rule(rule)
        if not self:
            self.add_rule(alt)


# ParenList {{{1
class ParenList(Production):
    """
    A list with parentheses as end caps and commas as the separator.

    :param items:
        A list of rules. Each will be joined together using the separator to 
        form a single unit.
    :param sep:
        A rule that is placed between each rule in rules.
    :param lcap:
        A rule that is placed before the first rule in rules.
    :param rcap:
        A rule that is placed after the last rule in rules.

    >>> from docgen import ParenList

    >>> args = ParenList(['a', 'b', 'c'])
    >>> print str(args)
    (a, b, c)

    """
    def __init__(self, items, sep=', ', lcap='(', rcap=')'):
        self.add_rule(Join(items, sep=sep, lcap=lcap, rcap=rcap))

# Required {{{1
class Required(Production):
    """A required rule.

    The rule is passed unchanged after assuring that it is not null. If it is 
    null, an exception is thrown.

    :param rule:
        The rule that this production evaluates to as long as this rule is 
        active. If this rule is not active, this production throws a ValueError 
        exception.

    >>> from docgen import Required

    >>> result = Required('Mr. Bill')
    >>> print str(result)
    Mr. Bill

    >>> result = Required(None)
    Traceback (most recent call last):
    ...
    ValueError: required production is null.

    """
    def __init__(self, rule):
        self.add_rule(rule)
        if not self:
            raise ValueError('required production is null.')


# StrikeLine {{{1
class StrikeLine(Production):
    """The current line of text is deleted.

    >>> from docgen import StrikeLine, Newline, Optional, Leader, Lines

    >>> class Comment(Production):
    ...     def __init__(self, lines):
    ...         self.add_rule([
    ...             '/*', Newline(),
    ...             Leader(Lines(lines), '*'),
    ...             Newline(), ' */'
    ...         ])

    >>> title = 'Monte Python'
    >>> description = 'A British surreal comedy group.'
    >>> comment = Comment([title, Optional(description, StrikeLine())])
    >>> print str(comment)
    /*
     * Monte Python
     * A British surreal comedy group.
     */

    >>> description = None
    >>> comment = Comment([title, Optional(description, StrikeLine())])
    >>> print str(comment)
    /*
     * Monte Python
     */

    """
    def __init__(self):
        self.add_rule(killCode())


# Tab {{{1
class Tab(Production):
    """Adds a horizontal tab to specified column.

    If current column is already beyond specified column, a newline is output 
    before the tab.

    :param column:
        The desired column of the tab stop.

    >>> from docgen import Tab, Production

    >>> result = Production(['Name:', Tab(8), 'Carlos Danger'])
    >>> print str(result)
    Name:   Carlos Danger

    >>> result = Production(['Name:', Tab(4), 'Carlos Danger'])
    >>> print str(result)
    Name:
        Carlos Danger

    """
    def __init__(self, column=None):
        if column is None:
            column = DEFAULT_TABSTOP
        self.add_rule(tabCode(column))



# *DocGen* is an reverse parser, or unparser. A parser converts text into an 
# structured internal representation. *DocGen* does the opposite. It converts 
# an structured internal representation into text. In other words, *DocGen* 
# converts data from data structures into text.

# Copyright (c) 2011-2015 by Designer's Guide Consulting, Inc.
# This code and documentation is the property of Designer's Guide Consulting and 
# is a corporate secret.  If you gain access to this code or documentation and 
# are not an employee of Designer's Guide Consulting, please notify 
# consulting@designers-guide.com immediately.  Until you hear back from them 
# please keep the code secret and avoid examining it. If you cannot reach them, 
# please destroy the code and documentation in a secure manner.

from .production import (
    Production,
    newlineCode, tabCode, killCode, emptyCode,
    indentCode, leaderCode, dedentCode, render
)
from .predefined import (
    Alternates, Block, Cull, DiscardIfEmpty, Indent, Indented, Join, Leader, 
    Lines, Newline, Op, Optional, ParenList, Required, StrikeLine, Tab
)
from .templet import Templet, TempletError
from .template import Template



# These settings for the Ply parse tables are placed in their own file so that 
# they can be easily replaced if this package is used in a larger package that 
# needs different settings.

import os

PREFIX = 'docgen'
LEXTAB = 'lextab'
PARSETAB = 'parsetab'
SUFFIX = str(os.getuid())
DIRECTORY = '/tmp'

def join(*components):
    return '_'.join([each for each in components if each])

class TabFiles(object):
    def __init__(self, filepath):
        root = os.path.splitext(filepath)[0]
        name = os.path.split(root)[1]
        self.lextab = join(PREFIX, name, LEXTAB, SUFFIX)
        self.parsetab = join(PREFIX, name, PARSETAB, SUFFIX)
        self.directory = DIRECTORY



# -*- coding: utf-8 -*-
'''
Provides the :class:`Arrow <arrow.arrow.Arrow>` class, an enhanced ``datetime``
replacement.

'''

from __future__ import absolute_import

from datetime import datetime, timedelta, tzinfo
from dateutil import tz as dateutil_tz
from dateutil.relativedelta import relativedelta
import calendar
import sys

from arrow import util, locales, parser, formatter


class Arrow(object):
    '''An :class:`Arrow <arrow.arrow.Arrow>` object.

    Implements the ``datetime`` interface, behaving as an aware ``datetime`` while implementing
    additional functionality.

    :param year: the calendar year.
    :param month: the calendar month.
    :param day: the calendar day.
    :param hour: (optional) the hour. Defaults to 0.
    :param minute: (optional) the minute, Defaults to 0.
    :param second: (optional) the second, Defaults to 0.
    :param microsecond: (optional) the microsecond. Defaults 0.
    :param tzinfo: (optional) the ``tzinfo`` object.  Defaults to ``None``.

    If tzinfo is None, it is assumed to be UTC on creation.

    Usage::

        >>> import arrow
        >>> arrow.Arrow(2013, 5, 5, 12, 30, 45)
        <Arrow [2013-05-05T12:30:45+00:00]>

    '''

    resolution = datetime.resolution

    _ATTRS = ['year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond']
    _ATTRS_PLURAL = ['{0}s'.format(a) for a in _ATTRS]
    _MONTHS_PER_QUARTER = 3

    def __init__(self, year, month, day, hour=0, minute=0, second=0, microsecond=0,
                 tzinfo=None):

        if util.isstr(tzinfo):
            tzinfo = parser.TzinfoParser.parse(tzinfo)
        tzinfo = tzinfo or dateutil_tz.tzutc()

        self._datetime = datetime(year, month, day, hour, minute, second,
            microsecond, tzinfo)


    # factories: single object, both original and from datetime.

    @classmethod
    def now(cls, tzinfo=None):
        '''Constructs an :class:`Arrow <arrow.arrow.Arrow>` object, representing "now".

        :param tzinfo: (optional) a ``tzinfo`` object. Defaults to local time.

        '''

        utc = datetime.utcnow().replace(tzinfo=dateutil_tz.tzutc())
        dt = utc.astimezone(dateutil_tz.tzlocal() if tzinfo is None else tzinfo)

        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, dt.tzinfo)

    @classmethod
    def utcnow(cls):
        ''' Constructs an :class:`Arrow <arrow.arrow.Arrow>` object, representing "now" in UTC
        time.

        '''

        dt = datetime.utcnow()

        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, dateutil_tz.tzutc())

    @classmethod
    def fromtimestamp(cls, timestamp, tzinfo=None):
        ''' Constructs an :class:`Arrow <arrow.arrow.Arrow>` object from a timestamp.

        :param timestamp: an ``int`` or ``float`` timestamp, or a ``str`` that converts to either.
        :param tzinfo: (optional) a ``tzinfo`` object.  Defaults to local time.

        '''

        tzinfo = tzinfo or dateutil_tz.tzlocal()
        timestamp = cls._get_timestamp_from_input(timestamp)
        dt = datetime.fromtimestamp(timestamp, tzinfo)

        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, tzinfo)

    @classmethod
    def utcfromtimestamp(cls, timestamp):
        '''Constructs an :class:`Arrow <arrow.arrow.Arrow>` object from a timestamp, in UTC time.

        :param timestamp: an ``int`` or ``float`` timestamp, or a ``str`` that converts to either.

        '''

        timestamp = cls._get_timestamp_from_input(timestamp)
        dt = datetime.utcfromtimestamp(timestamp)

        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, dateutil_tz.tzutc())

    @classmethod
    def fromdatetime(cls, dt, tzinfo=None):
        ''' Constructs an :class:`Arrow <arrow.arrow.Arrow>` object from a ``datetime`` and optional
        ``tzinfo`` object.

        :param dt: the ``datetime``
        :param tzinfo: (optional) a ``tzinfo`` object.  Defaults to UTC.

        '''

        tzinfo = tzinfo or dt.tzinfo or dateutil_tz.tzutc()

        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, tzinfo)

    @classmethod
    def fromdate(cls, date, tzinfo=None):
        ''' Constructs an :class:`Arrow <arrow.arrow.Arrow>` object from a ``date`` and optional
        ``tzinfo`` object.  Time values are set to 0.

        :param date: the ``date``
        :param tzinfo: (optional) a ``tzinfo`` object.  Defaults to UTC.
        '''

        tzinfo = tzinfo or dateutil_tz.tzutc()

        return cls(date.year, date.month, date.day, tzinfo=tzinfo)

    @classmethod
    def strptime(cls, date_str, fmt, tzinfo=None):
        ''' Constructs an :class:`Arrow <arrow.arrow.Arrow>` object from a date string and format,
        in the style of ``datetime.strptime``.

        :param date_str: the date string.
        :param fmt: the format string.
        :param tzinfo: (optional) an optional ``tzinfo``
        '''

        dt = datetime.strptime(date_str, fmt)
        tzinfo = tzinfo or dt.tzinfo

        return cls(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, tzinfo)


    # factories: ranges and spans

    @classmethod
    def range(cls, frame, start, end=None, tz=None, limit=None):
        ''' Returns an array of :class:`Arrow <arrow.arrow.Arrow>` objects, representing
        an iteration of time between two inputs.

        :param frame: the timeframe.  Can be any ``datetime`` property (day, hour, minute...).
        :param start: A datetime expression, the start of the range.
        :param end: (optional) A datetime expression, the end of the range.
        :param tz: (optional) A timezone expression.  Defaults to UTC.
        :param limit: (optional) A maximum number of tuples to return.

        **NOTE**: the **end** or **limit** must be provided.  Call with **end** alone to
        return the entire range, with **limit** alone to return a maximum # of results from the
        start, and with both to cap a range at a maximum # of results.

        Supported frame values: year, quarter, month, week, day, hour, minute, second

        Recognized datetime expressions:

            - An :class:`Arrow <arrow.arrow.Arrow>` object.
            - A ``datetime`` object.

        Recognized timezone expressions:

            - A ``tzinfo`` object.
            - A ``str`` describing a timezone, similar to 'US/Pacific', or 'Europe/Berlin'.
            - A ``str`` in ISO-8601 style, as in '+07:00'.
            - A ``str``, one of the following:  'local', 'utc', 'UTC'.

        Usage:

            >>> start = datetime(2013, 5, 5, 12, 30)
            >>> end = datetime(2013, 5, 5, 17, 15)
            >>> for r in arrow.Arrow.range('hour', start, end):
            ...     print repr(r)
            ...
            <Arrow [2013-05-05T12:30:00+00:00]>
            <Arrow [2013-05-05T13:30:00+00:00]>
            <Arrow [2013-05-05T14:30:00+00:00]>
            <Arrow [2013-05-05T15:30:00+00:00]>
            <Arrow [2013-05-05T16:30:00+00:00]>

        '''

        _, frame_relative, relative_steps = cls._get_frames(frame)

        tzinfo = cls._get_tzinfo(start.tzinfo if tz is None else tz)

        start = cls._get_datetime(start).replace(tzinfo=tzinfo)
        end, limit = cls._get_iteration_params(end, limit)
        end = cls._get_datetime(end).replace(tzinfo=tzinfo)

        current = cls.fromdatetime(start)
        results = []

        while current <= end and len(results) < limit:
            results.append(current)

            values = [getattr(current, f) for f in cls._ATTRS]
            current = cls(*values, tzinfo=tzinfo) + relativedelta(**{frame_relative: relative_steps})

        return results


    @classmethod
    def span_range(cls, frame, start, end, tz=None, limit=None):
        ''' Returns an array of tuples, each :class:`Arrow <arrow.arrow.Arrow>` objects,
        representing a series of timespans between two inputs.

        :param frame: the timeframe.  Can be any ``datetime`` property (day, hour, minute...).
        :param start: A datetime expression, the start of the range.
        :param end: (optional) A datetime expression, the end of the range.
        :param tz: (optional) A timezone expression.  Defaults to UTC.
        :param limit: (optional) A maximum number of tuples to return.

        **NOTE**: the **end** or **limit** must be provided.  Call with **end** alone to
        return the entire range, with **limit** alone to return a maximum # of results from the
        start, and with both to cap a range at a maximum # of results.

        Supported frame values: year, quarter, month, week, day, hour, minute, second

        Recognized datetime expressions:

            - An :class:`Arrow <arrow.arrow.Arrow>` object.
            - A ``datetime`` object.

        Recognized timezone expressions:

            - A ``tzinfo`` object.
            - A ``str`` describing a timezone, similar to 'US/Pacific', or 'Europe/Berlin'.
            - A ``str`` in ISO-8601 style, as in '+07:00'.
            - A ``str``, one of the following:  'local', 'utc', 'UTC'.

        Usage:

            >>> start = datetime(2013, 5, 5, 12, 30)
            >>> end = datetime(2013, 5, 5, 17, 15)
            >>> for r in arrow.Arrow.span_range('hour', start, end):
            ...     print r
            ...
            (<Arrow [2013-05-05T12:00:00+00:00]>, <Arrow [2013-05-05T12:59:59.999999+00:00]>)
            (<Arrow [2013-05-05T13:00:00+00:00]>, <Arrow [2013-05-05T13:59:59.999999+00:00]>)
            (<Arrow [2013-05-05T14:00:00+00:00]>, <Arrow [2013-05-05T14:59:59.999999+00:00]>)
            (<Arrow [2013-05-05T15:00:00+00:00]>, <Arrow [2013-05-05T15:59:59.999999+00:00]>)
            (<Arrow [2013-05-05T16:00:00+00:00]>, <Arrow [2013-05-05T16:59:59.999999+00:00]>)

        '''
        tzinfo = cls._get_tzinfo(start.tzinfo if tz is None else tz)
        start = cls.fromdatetime(start, tzinfo).span(frame)[0]
        _range = cls.range(frame, start, end, tz, limit)
        return [r.span(frame) for r in _range]


    # representations

    def __repr__(self):

        dt = self._datetime
        attrs = ', '.join([str(i) for i in [dt.year, dt.month, dt.day, dt.hour, dt.minute,
            dt.second, dt.microsecond]])

        return '<{0} [{1}]>'.format(self.__class__.__name__, self.__str__())

    def __str__(self):
        return self._datetime.isoformat()

    def __format__(self, formatstr):

        if len(formatstr) > 0:
            return self.format(formatstr)

        return str(self)

    def __hash__(self):
        return self._datetime.__hash__()


    # attributes & properties

    def __getattr__(self, name):

        if name == 'week':
            return self.isocalendar()[1]

        if name == 'quarter':
            return int(self.month/self._MONTHS_PER_QUARTER) + 1

        if not name.startswith('_'):
            value = getattr(self._datetime, name, None)

            if value is not None:
                return value

        return object.__getattribute__(self, name)

    @property
    def tzinfo(self):
        ''' Gets the ``tzinfo`` of the :class:`Arrow <arrow.arrow.Arrow>` object. '''

        return self._datetime.tzinfo

    @tzinfo.setter
    def tzinfo(self, tzinfo):
        ''' Sets the ``tzinfo`` of the :class:`Arrow <arrow.arrow.Arrow>` object. '''

        self._datetime = self._datetime.replace(tzinfo=tzinfo)

    @property
    def datetime(self):
        ''' Returns a datetime representation of the :class:`Arrow <arrow.arrow.Arrow>` object. '''

        return self._datetime

    @property
    def naive(self):
        ''' Returns a naive datetime representation of the :class:`Arrow <arrow.arrow.Arrow>` object. '''

        return self._datetime.replace(tzinfo=None)

    @property
    def timestamp(self):
        ''' Returns a timestamp representation of the :class:`Arrow <arrow.arrow.Arrow>` object. '''

        return calendar.timegm(self._datetime.utctimetuple())

    @property
    def float_timestamp(self):
        ''' Returns a floating-point representation of the :class:`Arrow <arrow.arrow.Arrow>` object. '''

        return self.timestamp + float(self.microsecond) / 1000000


    # mutation and duplication.

    def clone(self):
        ''' Returns a new :class:`Arrow <arrow.arrow.Arrow>` object, cloned from the current one.

        Usage:

            >>> arw = arrow.utcnow()
            >>> cloned = arw.clone()

        '''

        return self.fromdatetime(self._datetime)

    def replace(self, **kwargs):
        ''' Returns a new :class:`Arrow <arrow.arrow.Arrow>` object with attributes updated
        according to inputs.

        Use single property names to set their value absolutely:

        >>> import arrow
        >>> arw = arrow.utcnow()
        >>> arw
        <Arrow [2013-05-11T22:27:34.787885+00:00]>
        >>> arw.replace(year=2014, month=6)
        <Arrow [2014-06-11T22:27:34.787885+00:00]>

        Use plural property names to shift their current value relatively:

        >>> arw.replace(years=1, months=-1)
        <Arrow [2014-04-11T22:27:34.787885+00:00]>

        You can also provide a timezone expression can also be replaced:

        >>> arw.replace(tzinfo=tz.tzlocal())
        <Arrow [2013-05-11T22:27:34.787885-07:00]>

        Recognized timezone expressions:

            - A ``tzinfo`` object.
            - A ``str`` describing a timezone, similar to 'US/Pacific', or 'Europe/Berlin'.
            - A ``str`` in ISO-8601 style, as in '+07:00'.
            - A ``str``, one of the following:  'local', 'utc', 'UTC'.

        '''

        absolute_kwargs = {}
        relative_kwargs = {}

        for key, value in kwargs.items():

            if key in self._ATTRS:
                absolute_kwargs[key] = value
            elif key in self._ATTRS_PLURAL or key in ['weeks', 'quarters']:
                relative_kwargs[key] = value
            elif key in ['week', 'quarter']:
                raise AttributeError('setting absolute {0} is not supported'.format(key))
            elif key !='tzinfo':
                raise AttributeError('unknown attribute: "{0}"'.format(key))

        # core datetime does not support quarters, translate to months.
        if 'quarters' in relative_kwargs.keys():
            if relative_kwargs.get('months') is None:
                relative_kwargs['months'] = 0
            relative_kwargs['months'] += (value * self._MONTHS_PER_QUARTER)
            relative_kwargs.pop('quarters')
            
        current = self._datetime.replace(**absolute_kwargs)
        current += relativedelta(**relative_kwargs)

        tzinfo = kwargs.get('tzinfo')

        if tzinfo is not None:
            tzinfo = self._get_tzinfo(tzinfo)
            current = current.replace(tzinfo=tzinfo)

        return self.fromdatetime(current)

    def to(self, tz):
        ''' Returns a new :class:`Arrow <arrow.arrow.Arrow>` object, converted to the target
        timezone.

        :param tz: an expression representing a timezone.

        Recognized timezone expressions:

            - A ``tzinfo`` object.
            - A ``str`` describing a timezone, similar to 'US/Pacific', or 'Europe/Berlin'.
            - A ``str`` in ISO-8601 style, as in '+07:00'.
            - A ``str``, one of the following:  'local', 'utc', 'UTC'.

        Usage::

            >>> utc = arrow.utcnow()
            >>> utc
            <Arrow [2013-05-09T03:49:12.311072+00:00]>

            >>> utc.to('US/Pacific')
            <Arrow [2013-05-08T20:49:12.311072-07:00]>

            >>> utc.to(tz.tzlocal())
            <Arrow [2013-05-08T20:49:12.311072-07:00]>

            >>> utc.to('-07:00')
            <Arrow [2013-05-08T20:49:12.311072-07:00]>

            >>> utc.to('local')
            <Arrow [2013-05-08T20:49:12.311072-07:00]>

            >>> utc.to('local').to('utc')
            <Arrow [2013-05-09T03:49:12.311072+00:00]>

        '''

        if not isinstance(tz, tzinfo):
            tz = parser.TzinfoParser.parse(tz)

        dt = self._datetime.astimezone(tz)

        return self.__class__(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second,
            dt.microsecond, dt.tzinfo)

    def span(self, frame, count=1):
        ''' Returns two new :class:`Arrow <arrow.arrow.Arrow>` objects, representing the timespan
        of the :class:`Arrow <arrow.arrow.Arrow>` object in a given timeframe.

        :param frame: the timeframe.  Can be any ``datetime`` property (day, hour, minute...).
        :param count: (optional) the number of frames to span.

        Supported frame values: year, quarter, month, week, day, hour, minute, second

        Usage::

            >>> arrow.utcnow()
            <Arrow [2013-05-09T03:32:36.186203+00:00]>

            >>> arrow.utcnow().span('hour')
            (<Arrow [2013-05-09T03:00:00+00:00]>, <Arrow [2013-05-09T03:59:59.999999+00:00]>)

            >>> arrow.utcnow().span('day')
            (<Arrow [2013-05-09T00:00:00+00:00]>, <Arrow [2013-05-09T23:59:59.999999+00:00]>)

            >>> arrow.utcnow().span('day', count=2)
            (<Arrow [2013-05-09T00:00:00+00:00]>, <Arrow [2013-05-10T23:59:59.999999+00:00]>)

        '''

        frame_absolute, frame_relative, relative_steps = self._get_frames(frame)

        if frame_absolute == 'week':
            attr = 'day'
        elif frame_absolute == 'quarter':
            attr = 'month'
        else:
            attr = frame_absolute

        index = self._ATTRS.index(attr)
        frames = self._ATTRS[:index + 1]

        values = [getattr(self, f) for f in frames]

        for i in range(3 - len(values)):
            values.append(1)

        floor = self.__class__(*values, tzinfo=self.tzinfo)

        if frame_absolute == 'week':
            floor = floor + relativedelta(days=-(self.isoweekday() - 1))
        elif frame_absolute == 'quarter':
            floor = floor + relativedelta(months=-((self.month - 1) % 3))

        ceil = floor + relativedelta(
            **{frame_relative: count * relative_steps}) + relativedelta(microseconds=-1)

        return floor, ceil

    def floor(self, frame):
        ''' Returns a new :class:`Arrow <arrow.arrow.Arrow>` object, representing the "floor"
        of the timespan of the :class:`Arrow <arrow.arrow.Arrow>` object in a given timeframe.
        Equivalent to the first element in the 2-tuple returned by
        :func:`span <arrow.arrow.Arrow.span>`.

        :param frame: the timeframe.  Can be any ``datetime`` property (day, hour, minute...).

        Usage::

            >>> arrow.utcnow().floor('hour')
            <Arrow [2013-05-09T03:00:00+00:00]>
        '''

        return self.span(frame)[0]

    def ceil(self, frame):
        ''' Returns a new :class:`Arrow <arrow.arrow.Arrow>` object, representing the "ceiling"
        of the timespan of the :class:`Arrow <arrow.arrow.Arrow>` object in a given timeframe.
        Equivalent to the second element in the 2-tuple returned by
        :func:`span <arrow.arrow.Arrow.span>`.

        :param frame: the timeframe.  Can be any ``datetime`` property (day, hour, minute...).

        Usage::

            >>> arrow.utcnow().ceil('hour')
            <Arrow [2013-05-09T03:59:59.999999+00:00]>
        '''

        return self.span(frame)[1]


    # string output and formatting.

    def format(self, fmt='YYYY-MM-DD HH:mm:ssZZ', locale='en_us'):
        ''' Returns a string representation of the :class:`Arrow <arrow.arrow.Arrow>` object,
        formatted according to a format string.

        :param fmt: the format string.

        Usage::

            >>> arrow.utcnow().format('YYYY-MM-DD HH:mm:ss ZZ')
            '2013-05-09 03:56:47 -00:00'

            >>> arrow.utcnow().format('X')
            '1368071882'

            >>> arrow.utcnow().format('MMMM DD, YYYY')
            'May 09, 2013'

            >>> arrow.utcnow().format()
            '2013-05-09 03:56:47 -00:00'

        '''

        return formatter.DateTimeFormatter(locale).format(self._datetime, fmt)


    def humanize(self, other=None, locale='en_us', only_distance=False):
        ''' Returns a localized, humanized representation of a relative difference in time.

        :param other: (optional) an :class:`Arrow <arrow.arrow.Arrow>` or ``datetime`` object.
            Defaults to now in the current :class:`Arrow <arrow.arrow.Arrow>` object's timezone.
        :param locale: (optional) a ``str`` specifying a locale.  Defaults to 'en_us'.
        :param only_distance: (optional) returns only time difference eg: "11 seconds" without "in" or "ago" part.
        Usage::

            >>> earlier = arrow.utcnow().replace(hours=-2)
            >>> earlier.humanize()
            '2 hours ago'

            >>> later = later = earlier.replace(hours=4)
            >>> later.humanize(earlier)
            'in 4 hours'

        '''

        locale = locales.get_locale(locale)

        if other is None:
            utc = datetime.utcnow().replace(tzinfo=dateutil_tz.tzutc())
            dt = utc.astimezone(self._datetime.tzinfo)

        elif isinstance(other, Arrow):
            dt = other._datetime

        elif isinstance(other, datetime):
            if other.tzinfo is None:
                dt = other.replace(tzinfo=self._datetime.tzinfo)
            else:
                dt = other.astimezone(self._datetime.tzinfo)

        else:
            raise TypeError()

        delta = int(util.total_seconds(self._datetime - dt))
        sign = -1 if delta < 0 else 1
        diff = abs(delta)
        delta = diff

        if diff < 10:
            return locale.describe('now', only_distance=only_distance)

        if diff < 45:
            return locale.describe('seconds', sign, only_distance=only_distance)

        elif diff < 90:
            return locale.describe('minute', sign, only_distance=only_distance)
        elif diff < 2700:
            minutes = sign * int(max(delta / 60, 2))
            return locale.describe('minutes', minutes, only_distance=only_distance)

        elif diff < 5400:
            return locale.describe('hour', sign, only_distance=only_distance)
        elif diff < 79200:
            hours = sign * int(max(delta / 3600, 2))
            return locale.describe('hours', hours, only_distance=only_distance)

        elif diff < 129600:
            return locale.describe('day', sign, only_distance=only_distance)
        elif diff < 2160000:
            days = sign * int(max(delta / 86400, 2))
            return locale.describe('days', days, only_distance=only_distance)

        elif diff < 3888000:
            return locale.describe('month', sign, only_distance=only_distance)
        elif diff < 29808000:
            self_months = self._datetime.year * 12 + self._datetime.month
            other_months = dt.year * 12 + dt.month

            months = sign * int(max(abs(other_months - self_months), 2))

            return locale.describe('months', months, only_distance=only_distance)

        elif diff < 47260800:
            return locale.describe('year', sign, only_distance=only_distance)
        else:
            years = sign * int(max(delta / 31536000, 2))
            return locale.describe('years', years, only_distance=only_distance)


    # math

    def __add__(self, other):

        if isinstance(other, (timedelta, relativedelta)):
            return self.fromdatetime(self._datetime + other, self._datetime.tzinfo)

        raise TypeError()

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):

        if isinstance(other, timedelta):
            return self.fromdatetime(self._datetime - other, self._datetime.tzinfo)

        elif isinstance(other, datetime):
            return self._datetime - other

        elif isinstance(other, Arrow):
            return self._datetime - other._datetime

        raise TypeError()

    def __rsub__(self, other):

        if isinstance(other, datetime):
            return other - self._datetime

        raise TypeError()


    # comparisons

    def _cmperror(self, other):
        raise TypeError('can\'t compare \'{0}\' to \'{1}\''.format(
            type(self), type(other)))

    def __eq__(self, other):

        if not isinstance(other, (Arrow, datetime)):
            return False

        other = self._get_datetime(other)

        return self._datetime == self._get_datetime(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):

        if not isinstance(other, (Arrow, datetime)):
            self._cmperror(other)

        return self._datetime > self._get_datetime(other)

    def __ge__(self, other):

        if not isinstance(other, (Arrow, datetime)):
            self._cmperror(other)

        return self._datetime >= self._get_datetime(other)

    def __lt__(self, other):

        if not isinstance(other, (Arrow, datetime)):
            self._cmperror(other)

        return self._datetime < self._get_datetime(other)

    def __le__(self, other):

        if not isinstance(other, (Arrow, datetime)):
            self._cmperror(other)

        return self._datetime <= self._get_datetime(other)


    # datetime methods

    def date(self):
        ''' Returns a ``date`` object with the same year, month and day. '''

        return self._datetime.date()

    def time(self):
        ''' Returns a ``time`` object with the same hour, minute, second, microsecond. '''

        return self._datetime.time()

    def timetz(self):
        ''' Returns a ``time`` object with the same hour, minute, second, microsecond and tzinfo. '''

        return self._datetime.timetz()

    def astimezone(self, tz):
        ''' Returns a ``datetime`` object, adjusted to the specified tzinfo.

        :param tz: a ``tzinfo`` object.

        '''

        return self._datetime.astimezone(tz)

    def utcoffset(self):
        ''' Returns a ``timedelta`` object representing the whole number of minutes difference from UTC time. '''

        return self._datetime.utcoffset()

    def dst(self):
        ''' Returns the daylight savings time adjustment. '''
        return self._datetime.dst()

    def timetuple(self):
        ''' Returns a ``time.struct_time``, in the current timezone. '''

        return self._datetime.timetuple()

    def utctimetuple(self):
        ''' Returns a ``time.struct_time``, in UTC time. '''

        return self._datetime.utctimetuple()

    def toordinal(self):
        ''' Returns the proleptic Gregorian ordinal of the date. '''

        return self._datetime.toordinal()

    def weekday(self):
        ''' Returns the day of the week as an integer (0-6). '''

        return self._datetime.weekday()

    def isoweekday(self):
        ''' Returns the ISO day of the week as an integer (1-7). '''

        return self._datetime.isoweekday()

    def isocalendar(self):
        ''' Returns a 3-tuple, (ISO year, ISO week number, ISO weekday). '''

        return self._datetime.isocalendar()

    def isoformat(self, sep='T'):
        '''Returns an ISO 8601 formatted representation of the date and time. '''

        return self._datetime.isoformat(sep)

    def ctime(self):
        ''' Returns a ctime formatted representation of the date and time. '''

        return self._datetime.ctime()

    def strftime(self, format):
        ''' Formats in the style of ``datetime.strptime``.

        :param format: the format string.

        '''

        return self._datetime.strftime(format)

    def for_json(self):
        '''Serializes for the ``for_json`` protocol of simplejson.'''
        return self.isoformat()

    # internal tools.

    @staticmethod
    def _get_tzinfo(tz_expr):

        if tz_expr is None:
            return dateutil_tz.tzutc()
        if isinstance(tz_expr, tzinfo):
            return tz_expr
        else:
            try:
                return parser.TzinfoParser.parse(tz_expr)
            except parser.ParserError:
                raise ValueError('\'{0}\' not recognized as a timezone'.format(
                    tz_expr))

    @classmethod
    def _get_datetime(cls, expr):

        if isinstance(expr, Arrow):
            return expr.datetime

        if isinstance(expr, datetime):
            return expr

        try:
            expr = float(expr)
            return cls.utcfromtimestamp(expr).datetime
        except:
            raise ValueError(
                '\'{0}\' not recognized as a timestamp or datetime'.format(expr))

    @classmethod
    def _get_frames(cls, name):

        if name in cls._ATTRS:
            return name, '{0}s'.format(name), 1

        elif name in ['week', 'weeks']:
            return 'week', 'weeks', 1
        elif name in ['quarter', 'quarters']:
            return 'quarter', 'months', 3

        raise AttributeError()

    @classmethod
    def _get_iteration_params(cls, end, limit):

        if end is None:

            if limit is None:
                raise Exception('one of \'end\' or \'limit\' is required')

            return cls.max, limit

        else:
            return end, sys.maxsize

    @staticmethod
    def _get_timestamp_from_input(timestamp):

        try:
            return float(timestamp)
        except:
            raise ValueError('cannot parse \'{0}\' as a timestamp'.format(timestamp))

Arrow.min = Arrow.fromdatetime(datetime.min)
Arrow.max = Arrow.fromdatetime(datetime.max)



# Template class
#
# This was an attempt to provide a subset of the capabilities of the Production 
# class in a for that was easier to use and understand. The general feeling is 
# that it is still to complicate. It has been superseded by the Templet class 
# and it is anticipated that this class will soon be deprecated.

# Imports {{{1
from __future__ import print_function, division
from .production import Production, CTRL_CODE
from .predefined import Cull, Join, Newline, StrikeLine, Tab
import collections
import re


# Globals {{{1
ATTRIBUTE_SEPARATOR = '|'
NULL_INDICATOR = CTRL_CODE + '0'

# Utilities {{{1
# isAttribute {{{2
IDENT = r'[a-z_][a-z0-9_]*'
ATTRIBUTE = r'\*?{0}(?:\.{0})*'.format(IDENT)
ATTRIBUTE_MATCHER = re.compile(r'\A\s*%s\s*\Z' % ATTRIBUTE, re.I)
def isAttribute(e):
    return bool(ATTRIBUTE_MATCHER.match(e))

# isConditional {{{2
CONDITIONAL_MATCHER = re.compile(
    r'\A\s*%s\s*\Z' % r'\s*'.join([IDENT, r'\?', IDENT, ':', IDENT]), re.I)
def isConditional(e):
    return bool(CONDITIONAL_MATCHER.match(e))

# isString {{{2
STRING = r'''("([^\\"]|(\\.))*")|('([^\\']|(\\.))*')'''
COMPILED_STRING = re.compile(r'''\A\s*%s\s*\Z''' % STRING)
def isString(e):
    return bool(COMPILED_STRING.match(e))

# isJoin {{{2
JOIN_MATCHER = re.compile(r'\A\s*(%s)\s+(%s)\s*\Z' % (ATTRIBUTE, STRING), re.I)
def isJoin(e):
    match = JOIN_MATCHER.match(e)
    if match:
        fields = match.groups()
        return fields[0], fields[1].strip('"' + "'")


# Template class {{{1
class Template(Production):
    # Description {{{2
    """A Description of a Syntax Fragment

    A Template takes a string as input, which is converted to a production rule.

    :param template:
        A string that may contain brace delimited regions. The contents of 
        the brace delimited regions is evaluated and the result is 
        interpolated into the string. Within the braces may be a simple 
        identifier or it may be a python expression. If it is an identifier, 
        it must be known by the production object itself. Otherwise it is 
        expected to be a python expression.  There are two ways of informing 
        the production object of values of attributes, or the meaning of 
        names passed to the python expression: they are either passed as 
        a dictionary to the symbols argument or they are passed as extra 
        keyword arguments to the constructor.

        Normally any simple attributes that are interpolated into a template 
        are not further processed. However, if the attribute name is 
        preceded by an asterisk, then the attribute value is itself treated 
        as a template and so any attributes it contains would be elaborated. 
        In this case, the values of these attributes must also be known by 
        the production object.

        There is more features available, see the users guide for more 
        information.

    :param symbols:
        A dictionary containing names that should be made available during 
        Python expression evaluation within templates.

    :param tabstop:
        Tab stop to use if a tab character is found in the template.

    :param nullify:
        Treat inaccessible attributes as if they were null.

    :param * args:
        Unnamed arguments should be either class instances or dictionaries and 
        will be searched for attribute values in the order in which they are 
        given.

    :param ** kwargs:
        Any otherwise unrecognized  keyword arguments are saved and their 
        values made available as brace-delimited attributes in templates.
    """

    # Constructor {{{2
    def __init__(self, template, *args, **kwargs):
        #self.__dict__.update(kwargs)
        self.add_rule(Template._elaborate(self, template, *args, **kwargs))

    # add_template {{{2
    def add_template(self, template, *args, **kwargs):
        self.rule = self._clean(
            Template._elaborate(self, template, *args, **kwargs))

    # _elaborate {{{2
    @staticmethod
    def _elaborate(rule, template, *args, **kwargs):
        if type(template) == str:
            return Template._expand(rule, template, *args, **kwargs)
        elif isinstance(template, collections.Iterable):
            for t in template:
                res = Template._expand(rule, t, *args, **kwargs)
                if res:
                    break
            return res
        raise AssertionError("Invalid type for template value")

    # _expand {{{2
    @staticmethod
    def _expand(self, template, *args, **kwargs):
        # Description {{{3
        # Takes a string that contains nested braces and splits it into a list 
        # of productions. Everything outside the lowest level braces is treated 
        # as simple strings.  Things within the first level braces may be 
        # attribute names, in which case they are interpolated into the output.  
        # For example, in:
        #
        #   template = "Hello {user}."
        #
        # {user} is replaced by the value of user where user is either a kwarg 
        # or a named value associated with an arg in args.
        #
        # It is also possible for the contents of the first level braces to be a
        # python expression that evaluates to things that can be contained in
        # productions (strings, lists, and productions). In this case it is
        # generally required to specify the symbols argument so items in the
        # expression can be found.
        #
        #   template = "/* Created: {str(datetime.date.today())} */"
        #
        # If the thing in the first level of braces is a python expression (it 
        # is not a simple identifier), then it is possible to have a second 
        # level of braces. This second level of braces may only contain 
        # attribute names.
        #
        #   template = dedent('''\
        #       always @(*) begin
        #           {Indent(Join(
        #               ['In[%s] = {in}%s;' % (i, i) for i in range(4)],
        #               sep=Newline()
        #           ))}
        #       end''')
        #
        # You escape braces by doubling or tripling them up. When multiple 
        # braces are encountered, they are passed through except one level of 
        # braces are removed. In this way, the previous example could be 
        # rewritten as:
        #
        #   template = dedent('''\
        #       always @(*) begin
        #           {Indent(Lines(
        #               ['In[{{0}}] = {in}{{0}};'.format(i) for i in range(4)]
        #           ))}
        #       end''')
        #
        # In this case {in} is replaced by the value of rule.in and {{0}} is 
        # replaced with {0}, and so if rule.in is left_in, then python is 
        # passed:
        #
        #   Indent(Lines(['In[{0}] = left_in{0};'.format(i) for i in range(4)))
        #
        # This is a staticmethod with self passed in because we need to apply it 
        # to user classes that are not children of Production (those classes 
        # would be expected to have docgen_templates).

        # Fields class {{{3
        class Fields(object):
            # This is a class that mimics a dictionary. It must mimic 
            # a dictionary because its objects are passed to Python's eval() 
            # function, which requires dictionaries, but it is a class so that 
            # we can access the values dynamically. When requested, we search 
            # through the keyword arguments (kwargs is a dict), the arguments 
            # (each may be a dict or an object) and the base rule (object) 
            # looking for the desired key.
            #
            # It is expected that getValue will try __getitem__ before 
            # __getattr__, so our attributes should not interfere with the 
            # users, but I added leading and trailing underscores just in case.
            def __init__(self, kwargs, args, parent):
                self._local_symbols_ = {}
                self._kwargs_ = kwargs
                self._args_ = args
                self._parent_ = parent

            def __getitem__(self, key):
                for arg in (
                   [self._local_symbols_, self._kwargs_] +
                   list(self._args_) +
                   [self._parent_]
                ):
                    try:
                        return getattr(arg, key)
                    except (AttributeError, TypeError):
                        pass
                    try:
                        return arg[key]
                    except (KeyError, AttributeError, TypeError):
                        pass
                raise KeyError('%s: not found.' % key)

            def __contains__(self, key):
                try:
                    self._get(key)
                    return True
                except KeyError:
                    return False

            def __setitem__(self, key, value):
                # This is needed because eval() needs to store values for 
                # temporary variables (like: ['x[%s]' % n for n in ['p', 'n']]) 
                self._local_symbols_[key] = value

        # mismatch {{{3
        def mismatch():
            raise ValueError("unbalanced braces in '%s'" % template)

        # too_many_levels {{{3
        def too_many_levels():
            raise ValueError(
                "too many levels of braces (> 2) in '%s'" % template)

        # expected_identifier {{{3
        def expected_identifier():
            raise ValueError("expected identifier for '%s'" % template)

        # outputString {{{3
        def outputString(text):
            if text:
                return [text]
            else:
                return []

        # getValue {{{3
        def getValue(collection, path):
            # the path may be string of names separated with dots, if so call 
            # getValue recursively to track down the desired value, otherwise 
            # just return the value associated with the name contained in path.

            def get(collection, name):
                # collection may be an object, in which case name is the name of 
                # an attribute, or it may be a dictionary, in which case name is 
                # a key. collection may also be a Field object, which purposely 
                # has no attributes and instead mimics a dictionary.
                #
                # try as a dictionary first, in this way any attributes of Field 
                # will not be accessed by accident.
                try:
                    return collection[name]
                except (KeyError, AttributeError, TypeError):
                    pass
                try:
                    return getattr(collection, name)
                except AttributeError:
                    pass

                # The value was not found.
                # Either an error can be generated or None can be returned.
                # Choose your poison ...
                if nullify:
                    return None
                else:
                    raise KeyError('%s: not found.' % name)

            if '.' in path:
                name, remainder = path.split('.', 1)
                return getValue(get(collection, name), remainder)
            else:
                value = get(collection, path)
                assert Production._eligibleToBeRule(value), (
                    '%s: bad type: %s' % (path, type(value))
                )
                return value

        # expandValue {{{3
        def expandValue(attr):
            # Given a name, find the value associated with that name and return 
            # it.  If the name is preceded by an asterisk treat it as 
            # a template, which causes any attributes it contains to be expanded 
            # as well.
            if attr[0] == '*':
                # first, make sure that attribute value is a production so that 
                # the conversion back to a string works without a hitch
                #     this is weak in that we do not pass the attribute values 
                #     into any productions we find here, thus these productions 
                #     act on the attribute names rather than the attribute 
                #     values, meaning that any None processing is defeated.
                template = Production(getValue(fields, attr[1:]))
                # now, convert value into a string that is treated as a template
                # in this way, any attributes it contains will be evaluated
                rule = Template(str(template), fields)
                return rule
            # simply return the value of the expression
            return getValue(fields, attr)

        # evalExpression {{{3
        def evalExpression(expr):
            """Evaluate Python expression"""
            # should pass sumbols as 2nd argument to eval (globals) and fields 
            # to 3rd argument (locals). But I believe there is a bug in eval 
            # that sometimes skip looking in the local symbol table for local 
            # variables, so add kwargs to symbols and pass that in for the 
            # global symbol table. It will not fix the problem completely, but 
            # does allow of a work around: if it is not working, pass the 
            # offending value in as a kwarg.
            env = dict(list(symbols.items()) + list(kwargs.items()))
            result = eval(expr, env, fields)
            return result

        # outputNewlines {{{3
        def outputNewlines(newlines):
            return [Newline(newlines-1)]

        # _expand {{{3
        result = []
        start = 0
        depth = 0
        null = False
        last = len(template)
        symbols = kwargs.pop('symbols', {})
        tabstop = kwargs.pop('tabstop', None)
        nullify = kwargs.pop('nullify', None)
        fields = Fields(kwargs, args, self)

        iterator = iter(range(last))
        string = ''
        expr = ''
        for i in iterator:
            if template[i] == '{':
                # found opening brace
                braces = 1
                remainder = template[i+1:]
                if remainder and remainder[0] == '{':
                    # found escaped brace
                    # braces are escaped by doubling or tripling them up
                    for char in remainder:
                        if char == '{':
                            braces += 1
                            next(iterator)
                        else:
                            break
                    # add what we have seen before the braces to value to
                    # output and then add the number of braces found minus one.
                    if depth == 0:
                        string += template[start:i] + (braces-1)*'{'
                        expr = ''
                    elif depth == 1:
                        expr += template[start:i] + (braces-1)*'{'
                    else:
                        too_many_levels()
                else:
                    # found single opening brace
                    if depth == 0:
                        # output what we have so far because we may be entering
                        # a python expression
                        result += outputString(string + template[start:i])
                        string = ''
                    elif depth == 1:
                        # this is the start of an attribute
                        # save what we already have of the expression
                        expr += template[start:i]
                    else:
                        too_many_levels()
                    depth += 1
                start = i + braces
            elif template[i] == '}':
                # found closing brace
                braces = 1
                remainder = template[i+1:]
                if remainder and remainder[0] == '}':
                    # found multiple closing braces
                    for char in remainder:
                        if char == '}':
                            braces += 1
                            next(iterator)
                        else:
                            break
                    if depth == 0:
                        string += template[start:i] + (braces-1)*'}'
                    elif depth == 1:
                        expr += template[start:i] + (braces-1)*'}'
                    else:
                        too_many_levels()
                else:
                    # found single closing brace
                    if depth == 0:
                        mismatch()
                    elif depth == 1:
                        expr += template[start:i]
                        for term in expr.split(ATTRIBUTE_SEPARATOR): # | split
                            if isAttribute(term):
                                rule = expandValue(term)
                            elif term == '!strike':
                                rule = StrikeLine()
                            elif term == '':
                                rule = ''
                            elif isString(term):
                                if NULL_INDICATOR in term:
                                    rule = None
                                else:
                                    rule = term[1:-1]
                            elif isConditional(term): # match ?: oper
                                cond, rest = term.split('?')
                                iftrue, iffalse = rest.split(':')
                                if fields[cond]:
                                    rule = expandValue(iftrue)
                                else:
                                    rule = expandValue(iffalse)
                            else:
                                join = isJoin(term) # match join {name 'str'}
                                if join:
                                    rule = Join(
                                        Cull(expandValue(join[0])), sep=join[1]
                                    )
                                else:
                                    # is python expression
                                    if NULL_INDICATOR in term:
                                        rule = None
                                    else:
                                        rule = evalExpression(term)
                            rule = Production._clean(rule)
                            if rule is not None:
                                result += [rule]
                                break
                        else:
                            null = True
                        expr = ''
                    elif depth == 2:
                        name = template[start:i]
                        if not isAttribute(name):
                            expected_identifier()
                        value = getValue(fields, name)
                        if value is None:
                            expr += NULL_INDICATOR
                        elif type(value) == str:
                            expr += value
                        elif Production._eligibleToBeRule(value):
                            expr += Production._clean(value).expand()
                        else:
                            raise AssertionError(
                                ' '.join([
                                    "Values interpolated into Python",
                                    "expressions must be strings.",
                                    "name=%s, type=%s" % (name, type(value))
                                ])
                            )
                    else:
                        too_many_levels()
                    depth -= 1
                start = i + braces
            elif template[i] == '\n' and depth == 0:
                newlines = 1
                remainder = template[i+1:]
                if remainder and remainder[0] == '\n':
                    # found multiple newlines
                    for char in remainder:
                        if char == '\n':
                            newlines += 1
                            next(iterator)
                        else:
                            break
                result += outputString(template[start:i])
                result += outputNewlines(newlines)
                start = i + newlines
            elif template[i] == '\t' and depth == 0:
                result += outputString(template[start:i])
                result += [Tab(tabstop)]
                start = i + 1

        if depth != 0:
            mismatch()

        if null:
            return None
        else:
            return result + [s for s in [string, template[start:]] if s]


